package AOSICFG;
use LWP::UserAgent;
use Digest::MD5 qw(md5 md5_base64);
use XML::Simple;
use Sys::Hostname;
use MIME::Base64;
use Net::LDAP;
use Encode;

write_ws_log('===== Configuration =====');
write_ws_log('-=- Debug level '.$config::aosi_debug_level.' -=-');

write_ws_log('-=- Reading LDAP password from '.$config::ldap_pwd_file.' -=-');
$ldap_pwd = _get_pwd($config::ldap_pwd_file);

write_ws_log('-=- Reading admin user list from '.$config::valid_user_file.' -=-');
@valid_users = get_list($config::valid_user_file);

write_ws_log('-=- Getting remote schema file from '.$config::schema_url.' -=-');
my $schema_ok = get_schema($config::schema_url, $config::remote_schema_file);
if($schema_ok == 0){
    open(XML, "$config::remote_schema_file") or die "$0: error opening $config::remote_schema_file for reading ($!)\n";
    my $md5 = Digest::MD5->new;
    $md5->addfile(*XML);
    my $remote_md5 = $md5->b64digest;
    close(XML);

    my $local_md5 = '';
    write_ws_log('-=- Remote schema file MD5: '.$remote_md5.' -=-');
    if(-e $config::local_schema_file){
        open(XML, "$config::local_schema_file") or die "$0: error opening $config::local_schema_file for reading ($!)\n";
        my $md5 = Digest::MD5->new;
        $md5->addfile(*XML);
        $local_md5 = $md5->b64digest;
        close(XML);
    }
    write_ws_log('-=- Local schema file MD5:  '.$local_md5.' -=-');

    if($local_md5 ne $remote_md5){
        write_ws_log('-=- Schema files are different:  '.$config::remote_schema_file.' => '.$config::local_schema_file.' -=-');
        if(-e $config::local_schema_file){
            rename($config::local_schema_file, $config::old_schema_file);
        }
        rename($config::remote_schema_file, $config::local_schema_file);
    }
}
else {
    $stg = "-=- Could not reach: ".$schema_ok."! ";
    if(!(-e $config::local_schema_file)){
        $stg .= "Local schema file does not exist! PANIC!!! -=-";
        die "=== Local schema file ".$config::local_schema_file." does not exist! ===\n";
    }
    else {
        $stg .= "Using local file ".$config::local_schema_file."! -=-";
    }
    write_ws_log($stg);
}
if(-e $config::local_schema_file){
    get_bin_attribs($config::local_schema_file, $config::bin_attrib_file);
    get_req_attribs($config::local_schema_file, $config::req_attrib_file);
}

write_ws_log('-=- Reading valid IP list from '.$config::valid_ip_file.' -=-');
@valid_ip = get_list($config::valid_ip_file);

write_ws_log('-=- Reading binary attribute list from '.$config::bin_attrib_file.' -=-');
@bin_attribs = get_list($config::bin_attrib_file);
@bin_attribs = map {chomp; $_;}  @bin_attribs;

write_ws_log('-=- Use valid_ip file for instant access: '.$config::use_valid_ip.' -=-');
write_ws_log('-=- Reading required attribute list from '.$config::req_attrib_file.' -=-');
@req_attribs = get_list($config::req_attrib_file);
@req_attribs = map {chomp; $_;}  @req_attribs;
write_ws_log("-=- Configuration done -=-");


sub get_bin_attribs {
    my($file_in, $file_out) = @_;
    my($xml, $att, $type, @bin, $att_name);

    $xml = XMLin($file_in, KeyAttr => [ ]);
    open(BIN, ">$config::bin_attrib_file") or die "$0: error opening $config::bin_attrib_file for writting ($!)\n";
    foreach $att(sort {$a->{'ldapname'} cmp $b->{'ldapname'}} @{$xml->{'attribute'}}){
        #print "att: ".$att->{'ldapname'}.";type: ";
        @keys = keys %{$att->{'type'}};
        $type = $keys[0];
        if($type =~ /binary/i){
            $att_name = $att->{'ldapname'};
            push(@bin, $att_name);
            print BIN $att_name,"\n";
        }
    }
    close(BIN);
    write_ws_log('-=- Writting binary attribute list ('.join(",", @bin).') to '.$config::bin_attrib_file.' -=-');
}

sub get_req_attribs {
    my($file_in, $file_out) = @_;
    my($att, $att_name, $type, $xml, @bin, @keys);

    $xml = XMLin($file_in, KeyAttr => [ ]);
    open(MUST, ">$config::req_attrib_file") or die "$0: error opening $config::req_attrib_file for writting ($!)\n";
    foreach $att(sort {$a->{'ldapname'} cmp $b->{'ldapname'}} @{$xml->{'attribute'}}){
        $must = $att->{'status'}->{'must'};
        if($must =~ /true/i){
            $att_name = $att->{'ldapname'};
            push(@req, $att_name);
            print MUST $att_name."\n";
        }
    }
    close(MUST);
    write_ws_log('-=- Writting required attribute list ('.join(",", @req).') to '.$config::req_attrib_file.' -=-');
}

sub get_schema {
    my($url, $file_out) = @_;
    my($ua, $response);

    $ua = LWP::UserAgent->new;
    $ua->timeout(10);
    $ua->env_proxy;

    $response = $ua->get($url);

    if ($response->is_success) {
        open(XML, ">$file_out") or die "($!)\n";
        print XML $response->content;
        close(XML);
        return 0;
    }
    else {
        return $response->code;
    }
}

sub timestamp {
    my($sec, $min, $hour, $mday, $mon, $year, $wday) = localtime(time);
    my(@mon_name);

    @mon_name = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
    return sprintf("%s %02d %02d:%02d:%02d", $mon_name[$mon], $mday, $hour, $min, $sec);
}

sub write_ws_log {
    my($stg) = @_;
    open(LOG, ">>$config::ws_log_file") or die "$0: can not open file $ws_log_file for appending ($!)\n";
    $stg =~ s/\n//gm;
    #$stg =~ s/\s\s+//g;
    print LOG timestamp()," ",hostname()," aosi[",$$,"]: ",utfize($stg),"\n";
    close(LOG);
}


sub write_pid {
    my($file) = @_;

    open(FILE, ">$file") or die "$0: can not open file $file for writting ($!)\n";
        print FILE $$;
    close(FILE);
}

sub _get_pwd {
    my($file) = @_;
    my($real);

    open(FILE, "$file") or die "$0: can not open file $file for reading ($!)\n";
    $real = <FILE>;
    close(FILE);
    chomp($real);
    return $real;
}


sub get_pwd {
    my($file) = @_;
    if($file eq $config::ldap_pwd_file){
        return _get_pwd($file);
    }
    else {
        return '';
    }
}

sub get_list {
    my($file) = @_;
    my(@real, $i);

    open(FILE, "$file") or die "$0: can not open file $file for reading ($!)\n";
    @real = <FILE>;
    close(FILE);

    chomp(@real);
    return @real;
}

sub user_split {
    my($in) = @_;
    my($uid, $realm);

    $uid = $in;
    if($uid =~ /\@/){
        ($uid, $realm) = split(/\@/, $uid);
    }
    return $uid, $realm;
}

sub _user_is_valid {
    my($in) = @_;
    my($ok, $realm, $uid, $user, @users);

    $ok = 0;
    ($uid, $realm) = user_split($in);
    foreach $user(@valid_users){
        chomp($user);
        if($user eq $uid){
            $ok++;
        }
    }
    write_ws_log('| User '.$in.' is valid: '.$ok) if $config::aosi_debug_level > 2;
    return $ok;
}

sub user_is_valid {
    my($in) = @_;
    return _user_is_valid($in);
}

sub _ip_is_valid {
    my($in) = @_;
    my($ip, $ok);

    $ok = 0;
    foreach $ip(@valid_ip){
        chomp($ip);
        if($ip eq $in){
            $ok++;
        }
    }
    write_ws_log('| IP '.$in.' is valid: '.$ok) if $config::aosi_debug_level > 2;
    return $ok;
}

sub ip_is_valid {
    my($in) = @_;
    return _ip_is_valid($in);
}

sub _is_bin {
    my($attrib) = @_;
    my($attr, $ok);

    $ok = 0;
    foreach $attr(@bin_attribs){
        chomp($attr);
        if($attr eq $attrib){
            $ok++;
        }
    }
    write_ws_log('| Attrib '.$attrib.' is binary: '.$ok) if $config::aosi_debug_level > 2;
    return $ok;
}

sub is_bin {
    my($attrib) = @_;
    return _is_bin($attrib);
}

sub dump_hash {
    my($hash, $dont_dump) = @_;
    my($out, $key, $val, $len);
    $out = '';

    foreach $key(keys %$hash){
        $val = $hash->{$key};
        if(ref($val) eq 'ARRAY'){
            $val = '['.join(',',@$val).']';                        	
        }
        $len = length($val);
        if($len > $config::log_large_value_limit){
            $val = '***Large value. Not dumped!***';
        }
        $out .= $key." => ".$val.";";

    }
    return $out;
}

sub dump_array {
    my($array, $dont_dump) = @_;
    my($out, $i, $val, $len);
    $out = '';

    foreach $i(0..$#{$array}){
        next if (($dont_dump eq 'defined') and ($dont_dump == $i));
        $val = $array->[$i];
        if(ref($val) eq 'ARRAY'){
            $val = '['.join(',',@$val).']';
        }
        $len = length($val);
        if($len > $config::log_large_value_limit){
            $val = '***Large value. Not dumped!***';
        }
        $out .= $val.";";

    }
    return $out;
}

sub utfize {
    my($in) = @_;

    if(utf8::is_utf8($in)){
            utf8::encode($in);
        }
    return $in;
}

######################################################################
# Intialize LDAP connection
######################################################################
sub ldap_init {
    my($hostname, $port)=@_;
    my($ldap, $ldap_error);

    $error = '';
    $ldap = Net::LDAP->new( $hostname,
        port => $port, debug => $config::ldap_debug_level, timeout => $config::ldap_timeout);
    unless($ldap){
        AOSICFG::write_ws_log("| Connection to LDAP server $hostname:$port failed: $@");
        $error = "Connection to LDAP server $hostname:$port failed: $@";
    }
    return $ldap, $error;
}

######################################################################
#
######################################################################
sub xml2entries {
    my($xml_in) = @_;
    my($attr, $attrs, $ent, $hash, $i, $key, $val, $xml, %attr_exists, @entries, @keys, @vals);

    AOSICFG::write_ws_log("|/ XML2Entries: xml_in=$xml_in") if $config::aosi_debug_level > 3;
    $xml = XMLin($xml_in, ForceArray => ['entry', 'value', 'attribute'], NormalizeSpace => 2);
    @entries=();
    foreach $ent(@{$xml->{'entry'}}){
        $hash = {};
        $attrs = undef;
        $hash->{'dn'} = $ent->{'dn'};
        AOSICFG::write_ws_log("|| XML2Entries: dn = ".$ent->{'dn'}."") if $config::aosi_debug_level > 3;
        $i=0;
        foreach $attr(@{$ent->{'attribute'}}){
            $key = $attr->{'ldapname'};
            $val = $attr->{'value'};
            # If the <value> is empty it is represented as 'value' => [ {} ]
            if(ref($val) eq 'ARRAY'){
                @vals = @$val;
                if($#vals == 0 and ref($vals[0]) eq 'HASH'){
                    @keys = keys %{$vals[0]};
                    if($#keys < 0){
                        $val = [];
                    }
                }
            }
            AOSICFG::write_ws_log("|| XML2Entries: value = (".join(", ", @$val).")") if $config::aosi_debug_level > 3;
            AOSICFG::write_ws_log("|| XML2Entries: att = $key; val = (".join(",", @$val).")") if $config::aosi_debug_level > 3;
            push(@$attrs, {'att' => $key, 'val' => $val, 'id' => $i});
            $i++;
            $attr_exists{$key}++;
        }
        $hash->{'attrs'} = $attrs;
        AOSICFG::write_ws_log("|| XML2Entries(hash):".AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 3;
        push(@entries, $hash);
    }
    AOSICFG::write_ws_log("|\\ XML2Entries End") if $config::aosi_debug_level > 3;
    return \@entries, \%attr_exists;
}


######################################################################
# Separate hostname and port part
######################################################################
sub parse_server {
    my($server) = @_;
    my($hostname, $port);

    if(index($server, ':') > -1){
        ($hostname, $port) = split(/:/, $server);
    }
    else {
        $hostname = $server;
        $port = 389;
    }
    return $hostname, $port;
}

######################################################################
#
######################################################################
sub get_dn {
    my($user, $base) = @_;
    my($uid, $realm, $dn);

    ($uid, $realm) = split(/\@/, $user);
    $dn = 'uid='.$uid.','.$base;
    return $dn;
}

######################################################################
#
######################################################################
sub get_uid_base {
    my($dn) = @_;
    my($uid, $realm, $dummy, $base, @base);

    AOSICFG::write_ws_log("|/ get_uid_base(in):dn:$dn;") if $config::aosi_debug_level > 3;
    ($uid, @base) = split(/,/, $dn);
    ($dummy, $uid) = split(/=/, $uid);
    $base = join(",", @base);
    AOSICFG::write_ws_log("|\\ get_uid_base(out):uid:$uid;base:$base;") if $config::aosi_debug_level > 3;
    return $uid, $base;
}

######################################################################
#
######################################################################
sub ldapGenSearch {
    my($service, $user, $pwd, $base, $filter, $attributes, $bind_user, $from, $size, $get_bin, $md5_in)=@_;
    my($auth_code, $auth_error, $code, $dec_pwd, $entry, $hash, $hostname);
    my($i, $is_admin, $ldap, $log, $md5, $mesg2, $out, $port, $show_bin, $to, $val, $attr, $cur_attr);
    my(%options, @attrs, @entries, @values, $admin_filter, $my_filter, $ldap_error, $dn);
    $hash = undef;
    $out = '';
    $show_bin = $get_bin || 0;
    $log = 'ldapGenSearch';

    AOSICFG::write_ws_log("| $log(in):user:$user;base:$base;filter:$filter;attrib:$attributes;user:$bind_user;md5:$md5_in;") if $config::aosi_debug_level > 2;

    ($hostname, $port) = AOSICFG::parse_server($config::ldap_server);
    ($ldap, $ldap_error) = ldap_init($hostname, $port);
    $dn = get_dn($user, $base);
    $admin_filter = '';
    if($bind_user == 1){
        ($auth_code, $auth_error) = AOSICFG::ck_auth($ldap, get_dn($user, $base), $pwd, 0, $ldap_error);
    }
    elsif($bind_user == -1){
        # Do nothing
        AOSICFG::write_ws_log("| $log:anonymous access");
        unless($ldap){
            $auth_code = -18;
            $auth_error = $ldap_error;
        }
    }
    else {
        ($auth_code, $auth_error, $is_admin, $admin_filter) = AOSICFG::ck_auth($ldap, get_dn($user, $base), $pwd, 1, $ldap_error);
        Encode::_utf8_on($admin_filter);
        if($is_admin == 0 and $ldap){
            if($auth_code == 0){
                $auth_code = -15;
                $auth_error = "User $user does not have admin privileges!";
            }
            if($from > -1){
               $hash->{'from'} = 0;
               $hash->{'size'} = 0;
               $hash->{'count'} = 0;
          }
        }
    }
    if($auth_code == 0){

      if($bind_user == 1){
          $dec_pwd = decode_base64($pwd);
          $mesg2 = $ldap->bind(get_dn($user, $base), password => $dec_pwd, version => 3);
          AOSICFG::write_ws_log("| $log(bind_as_user code): ".$mesg2->code) if $config::aosi_debug_level > 2;

      }
      elsif($bind_user == -1){
          $mesg2 = $ldap->bind(version => 3);
          AOSICFG::write_ws_log("| $log(bind_as_anonymous code): ".$mesg2->code) if $config::aosi_debug_level > 2;

      }
      else {
          $mesg2 = $ldap->bind($config::ldap_user, password => AOSICFG::get_pwd($config::ldap_pwd_file), version => 3);
          AOSICFG::write_ws_log("| $log(bind_as_admin code): ".$mesg2->code) if $config::aosi_debug_level > 2;

      }
      if($filter =~ /^\(.*\)$/){
            $my_filter = '(&'.$filter.''.$admin_filter.')';
        }
        else {
            $my_filter = '(&('.$filter.')'.$admin_filter.')';
        }
      AOSICFG::write_ws_log("| $log(my_filter): ".$my_filter) if $config::aosi_debug_level > 2;
      %options = (base   => $base,
          filter => $my_filter,
          #sizelimit => 0,
          timelimit => 180
          );
      if($attributes){
                @attrs = split(/[,; ]/, $attributes);
                $options{'attrs'} = \@attrs;
      }
      $mesg2 = $ldap->search(%options);
      $code = $mesg2->code;
      $hash->{'code'} = $code;
      $to = $from + $size - 1;
      $i = -1;
      if($code == 0){
          @entries = $mesg2->entries;
          LOOP:
          foreach $entry (sort bydn @entries) {
              if($from > -1){
                  $i++;
                  if($i < $from){
                        next LOOP;
                  }
                  if($i > $to){
                    last;
                  }
              }
              $dn = $entry->dn();
              $out .= "<entry";
              if($from > -1){
                  $out .= " num=\"$i\"";
              }
              $out .= " dn=\"$dn\">";
              foreach $attr ( $entry->attributes ) {
                  $cur_attr = $attr eq 'homePhone' ? 'homeTelephoneNumber' : $attr;
                  AOSICFG::write_ws_log("| $log(attr): $attr") if $config::aosi_debug_level > 3;
                  #$hash->{$attr} = join(', ', @{$entry->get_value( $attr, asref => 1 )});
                  $out .= "<attribute ldapname=\"".$cur_attr."\">";
                  @values = @{$entry->get_value( $attr, asref => 1 )};
                    if(AOSICFG::is_bin($attr)){
                        if($md5_in eq ''){
                            foreach $val(@values){
                                if($show_bin == 1){
                                    $out .= "<value>".encode_base64($val)."</value>";
                                }
                                else {
                                    $md5 = md5_base64($val);
                                    $out .= "<value>$md5</value>";
                                }
                            }
                        }
                        else {
                            foreach $val(@values){
                                $md5 = md5_base64($val);
                                #AOSICFG::write_ws_log("| ldapGenSearch(multi): md5:$md5;");
                                if($md5_in eq $md5){
                                    $out .= "<value>".encode_base64($val)."</value>";
                                }
                            }
                        }
                    }
                    else {
                        foreach $val(@values){
                            $out .= "<value>".$val."</value>";
                        }
                    }
                  $out .= "</attribute>";
              }
              AOSICFG::write_ws_log("| $log: XML building finished") if $config::aosi_debug_level > 3;
              $out .= "</entry>";
          }
          if($from > -1){
               $hash->{'from'} = $from;
               $hash->{'size'} = $size;
               $hash->{'count'} = $#entries + 1;
          }
          $out = "<?xml version=\"1.0\"?><ldap action=\"$service\" version=\"".$config::aosi_version."\">$out</ldap>";
          $hash->{'result'} = $out;
          $hash->{'error'} = '';
      }
      else {
          $hash->{'error'} = $mesg2->error;
          $hash->{'result'} = '';
      }
      $mesg2 = $ldap->unbind;  # take down session

    }
    else {
        $hash->{'code'} = $auth_code;
        $hash->{'error'} = $auth_error;
        $hash->{'result'} = '<?xml version="1.0"?><ldap action="'.$service.'" version="'.$config::aosi_version.'"><entry dn="'.$dn.'"><status code="'.$auth_code.'">'.$auth_error.'</status></entry></ldap>';
    }
    #$hash_val ='';
    AOSICFG::write_ws_log("| $log(out):".AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 2;
    return $hash;
}

sub bydn {
    $a->dn() cmp $b->dn();
}


######################################################################
# Generic subroutine for attribute manipulation
######################################################################
sub ldapGenModify {
    my($service, $user, $pwd, $base, $xml_in, $type, $bind_user) = @_;
    my($att, $attr_exists, $attrs, $auth_code, $auth_error, $bin_attr_ex);
    my($bin_ent, $bin_error, $bin_hash, $bin_msg, $bin_val, $bin_xml, $code, $cur_code, $cur_md5);
    my($cur_old_md5, $cur_val, $dec_pwd, $dn, $dn_in, $dummy, $ent, $entries, $err_msg, $err_num);
    my($hash, $hostname, $i, $is_admin, $j, $lbl, $ldap, $len, $log, $md5, $mesg2);
    my($mod_hash, $old, $old_bin, $old_val, $out_msg, $port, $prev_val, $realm, $temp_msg);
    my($uid, $val, %md5_del, %md5_exists, @del, @entries, @mod_keys, @new_vals, @old_md5);
    my(@old_vals, @to_del, @vals, @hep_af_vals, %options, $tmp_val);
    my($admin_filter, $my_filter, @entries2, $mesg3, $ldap_error);
    $log = 'ldapGenModify';
    AOSICFG::write_ws_log("| $log(in):user:$user;base:$base;xml:$xml_in;type:$type;bind_user:$bind_user") if $config::aosi_debug_level > 2;

    ($hostname, $port) = parse_server($config::ldap_server);
    ($ldap, $ldap_error) = ldap_init($hostname, $port);
    ($uid, $realm) = split(/\@/, $user);
    $dn = 'uid='.$uid.','.$base;
    $code = 0;
    $err_num = 0;
    $out_msg = '';

    if($bind_user == 1){
        ($auth_code, $auth_error) = ck_auth($ldap, get_dn($user, $base), $pwd, 0, $ldap_error);
    }
    else {
        ($auth_code, $auth_error, $is_admin, $admin_filter) = ck_auth($ldap, get_dn($user, $base), $pwd, 1, $ldap_error);
        Encode::_utf8_on($admin_filter);
        if($is_admin == 0 and $ldap){
            if($auth_code == 0){
                $auth_code = -15;
                $auth_error = "User $user does not have admin privileges!";
            }
        }

    }
    if($auth_code == 0){
        if($bind_user == 1){
            $dec_pwd = decode_base64($pwd);
            $dn = get_dn($user, $base);
            $mesg2 = $ldap->bind($dn, password => $dec_pwd, version => 3);
            AOSICFG::write_ws_log("| $log(bind_as_user): ".$mesg2->code) if $config::aosi_debug_level > 2;
        }
        else {
            $mesg2 = $ldap->bind($config::ldap_user, password => AOSICFG::get_pwd($config::ldap_pwd_file), version => 3);
            AOSICFG::write_ws_log("| $log(bind_as_admin): ".$mesg2->code) if $config::aosi_debug_level > 2;
        }
            ($entries, $attr_exists) = xml2entries($xml_in);
            @entries = @{$entries};


            foreach $ent(@entries){
                $dn_in = AOSICFG::utfize($ent->{'dn'});
                ($uid, $dummy) = split(/,/, $dn_in);
                ($lbl, $uid) = split(/=/, $uid);
                $out_msg .= '<entry dn="'.$dn_in.'">';
                $my_filter = '(&(uid='.$uid.')'.$admin_filter.')';
                AOSICFG::write_ws_log("| $log:my_filter:$my_filter") if $config::aosi_debug_level > 2;
                $mesg3 = $ldap->search(base => $base, filter => $my_filter, timelimit => 60,  attrs => ['uid']);
                $cur_code = $mesg3->code;
                @entries2 = $mesg3->entries;
                AOSICFG::write_ws_log("| $log(search with my_filter):code:$cur_code;error:".$mesg3->error.";entries:".$#entries2) if $config::aosi_debug_level > 2;
                if($cur_code == 0 and $#entries2 > -1){

                    $attrs = $ent->{'attrs'};
                    $len = $#{$attrs};
                    #@del = ();
                    #@vals = ();
                    $mod_hash = undef;
                    foreach $i(0..$len){
                        @del = ();
                        @vals = ();
                        $att = $attrs->[$i]->{'att'};
                        if($att eq  'phoneNumber'){
                            $att = 'homeTelephoneNumber';
                        }
                        $val = $attrs->[$i]->{'val'};
                        if(ref($val) eq 'ARRAY'){
                            @vals = @$val;
                            if($#vals == -1){
                                $vals[0] = '';
                            }
                            AOSICFG::write_ws_log("| $log($att; multi:".$#vals."): ".join(", ", @vals)) if $config::aosi_debug_level > 2;
                        }
                        else {
                            if($val){
                                push(@vals, $val);
                            }
                            AOSICFG::write_ws_log("| $log($att; single): $val")  if $config::aosi_debug_level > 2;
                        }
                        if(AOSICFG::is_bin($att)){
                            if($bind_user == 1){
                                #$bin_hash = AOSI::userBinSearch($service, $user, $pwd, $base, "(uid=$uid)", $att, undef, 1);
                                $bin_hash = ldapGenSearch($service, $user, $pwd, $base, "(uid=$uid)", $att, 1, -1, 0, 1, $md5)
                            }
                            else {
                                #$bin_hash = AOSI::ldapBinSearch($service, $user, $pwd, $base, "(uid=$uid)", $att, undef, 1);
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", $att, 0, -1, 0, 1, $md5);
                            }
    
                            $bin_xml = $bin_hash->{'result'};
                            ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                            $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                            if(ref($bin_val) eq 'ARRAY'){ # Multi
                                @old_vals = @$bin_val;
                            }
                            else { # Single
                                if($bin_val){
                                    @old_vals = ($bin_val);
                                }
                            }
                            AOSICFG::write_ws_log("| $log(old_multi:".$#old_vals."): ".join(", ", @old_vals)) if $config::aosi_debug_level > 2;
                            if($type == 1) { # Add
                                AOSICFG::write_ws_log("| $log(bin_add): ") if $config::aosi_debug_level > 2;
                                AOSICFG::write_ws_log("| $log(old_vals): ".$#old_vals) if $config::aosi_debug_level > 2;
                                if($#old_vals > -1){
                                    @old_md5 = ();
                                    foreach $old_val(@old_vals){
                                        $old_bin = decode_base64($old_val);
                                        $cur_old_md5 = md5_base64($old_val);
                                        $md5_exists{$cur_old_md5}++;
                                        push(@{$mod_hash->{$att}}, $old_bin);
                                    }
                                }
                                AOSICFG::write_ws_log("| $log(vals): ".$#vals) if $config::aosi_debug_level > 2;
                                foreach $cur_val(@vals){
                                    $cur_md5 = md5_base64($cur_val);
                                    if(!$md5_exists{$cur_md5}){
                                        push(@{$mod_hash->{$att}}, decode_base64($cur_val));
                                    }
                                }
                                $bin_error = 0;                            
                            }
                            elsif($type == 2) { # Modify
                                AOSICFG::write_ws_log("| $log(mod_bin): $md5") if $config::aosi_debug_level > 2;
                                $bin_error = -31;
                                $bin_msg = "Modification of binary attribute $att not allowed!";
                            }
                            elsif($type == 3) { # Delete
                                AOSICFG::write_ws_log("| $log(del_bin): $md5") if $config::aosi_debug_level > 2;
                                AOSICFG::write_ws_log("| $log(bin_vals): (".join(",", @vals).")") if $config::aosi_debug_level > 2;
                                foreach $val(@vals){
                                    $md5_del{$val}++;
                                }
                                if($#old_vals > -1 and $#vals > -1){
                                    @old_md5 = ();
                                    foreach $old_val(@old_vals){
                                        $old_bin = decode_base64($old_val);
                                        $cur_old_md5 = md5_base64($old_bin);
                                        push(@old_md5, $cur_old_md5);
                                        if(!$md5_del{$cur_old_md5}){
                                            push(@{$mod_hash->{$att}}, $old_bin);
                                        }
                                    }
                                    AOSICFG::write_ws_log("| $log(old_md5):  (".join(",", @old_md5).")") if $config::aosi_debug_level > 2;
                                }
                                $bin_error = 0;
                                AOSICFG::write_ws_log("| $log(mod_hash2):".AOSICFG::dump_hash($mod_hash))if $config::aosi_debug_level > 2;
                                #AOSICFG::write_ws_log("| $log(del_bin): $val");
                                #$mesg2 = $ldap->modify( $dn_in,  delete => [$att]);
                            }
                        }
                        else {
                            AOSICFG::write_ws_log("| $log($att; not_bin): ".join(", ", @vals)) if $config::aosi_debug_level > 2;
                            if($#vals > -1){
                                foreach $j(0..$#vals){
                                    $val = $vals[$j];
                                    AOSICFG::write_ws_log("| $log(cur_val): $val") if $config::aosi_debug_level > 2;
                                    if($val eq '' and $type == 3){
                                        push(@del, $att);
                                        AOSICFG::write_ws_log("| $log($att; del_all_1)") if $config::aosi_debug_level > 2;
                                    }
                                    else {
                                        #if($j == 0){
                                        #    $mod_hash->{$att} = $val;
                                        #}
                                        #elsif ($j == 1){
                                        #    $prev_val = $mod_hash->{$att};
                                        #    $mod_hash->{$att} = [$prev_val, $val];
                                        #}
                                        #elsif($j > 1){
                                        #    push(@{$mod_hash->{$att}}, $val);
                                        #}
                                        push(@{$mod_hash->{$att}}, $val);
                                        AOSICFG::write_ws_log("| $log(mod_not_bin):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                    }
                                    #$out_msg .= "$att;";
                                }
                            }
                            else {
                                if($type == 3){
                                    push(@del, $att);
                                    AOSICFG::write_ws_log("| $log($att; del_all_2)") if $config::aosi_debug_level > 2;
                                }
                            }
                        }
                    }
                    push(@$attrs, objectClass => 'hrEduPerson');
                    if(AOSICFG::is_bin($att)){
                        #$out_msg = '';
                        if(!$bin_error){
                            if($type == 1){ # Add
                                if($#old_vals > -1){
                                    $mesg2 = $ldap->modify( $dn_in,  delete => [$att]);
                                    AOSICFG::write_ws_log("| $log(bin_deleting 1aaa): dn=$dn_in;att=$att") if $config::aosi_debug_level > 2;
                                    AOSICFG::write_ws_log("| $log(bin_deleting 1): $att=".$mesg2->error."(".$mesg2->code.")") if $config::aosi_debug_level > 2;
                                    $cur_code = $mesg2->code;
                                }
                                else {
                                    AOSICFG::write_ws_log("| $log(bin_deleting 1a): $att=nothing to delete (old_vals:".$#old_vals.")") if $config::aosi_debug_level > 2;
                                    $cur_code = 0;
                                }
                                if($cur_code == 0){
                                    AOSICFG::write_ws_log("| $log(bin_adding a): $dn_in") if $config::aosi_debug_level > 2;
                                    AOSICFG::write_ws_log("| $log(bin_adding b):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                    $mesg2 = $ldap->modify( $dn_in,  add => $mod_hash);
                                    $cur_code = $mesg2->code;
                                    my $all_attrs = join(", ", keys %$mod_hash);
                                    AOSICFG::write_ws_log("| $log(bin_adding): $att=".$mesg2->error."(".$code.")") if $config::aosi_debug_level > 2;
                                    if($cur_code == 0){
                                        $tmp_msg = '<status code="0">Added binary: '.$all_attrs.'!</status>';
                                    }
                                    else {
                                        $tmp_msg = '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                        $err_num++;
                                    }
                                }
                                else {
                                    $tmp_msg = '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                    $err_num++;
                                }
                                $out_msg .= $tmp_msg;
                            }
                            elsif($type == 3){ # delete
                                @to_del = ();
                                foreach $old(@old_md5){
                                    if($md5_del{$old}){
                                        push(@to_del, $old);
                                    }
                                }
                                if($#to_del > -1){
                                    $mesg2 = $ldap->modify( $dn_in,  delete => [$att]);
                                    AOSICFG::write_ws_log("| $log(deleting): $att=".$mesg2->code);
                                    $out_msg .= '<status code="0">Deleted binary: $att!</status>';
                                    if($mod_hash->{$att}){
                                        $mesg2 = $ldap->modify( $dn_in,  add => $mod_hash);
                                        AOSICFG::write_ws_log("| $log(modifying): $att=".$mesg2->code);
                                    }
                                }
                                else {
                                    $bin_error = -32;
                                    $bin_msg = "Nothing to delete!";
                                }
                            }
                        }
                    }
                    else {
                        if($att eq 'homePostalAddress' or $att eq 'facsimileTelephoneNumber'){
                            if($bind_user == 1){
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", $att, 1, -1, 0);
                            }
                            else {
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", $att, 0, -1, 0);
                            }
                            AOSICFG::write_ws_log("| $log(hPA_old_vals):".AOSICFG::dump_hash($bin_hash)) if $config::aosi_debug_level > 2;
                            $bin_xml = $bin_hash->{'result'};
                            ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                            $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                            if($bin_val){
                                @old_vals = @$bin_val;
                            }
                            else {
                                @old_vals = ();
                            }
                            if($type == 1){  # Add
                                foreach $old_val(@old_vals){
                                    push(@new_vals, $old_val);
                                }
                                foreach $cur_val(@{$mod_hash->{$att}}){
                                    push(@new_vals, $cur_val);
                                }
                                AOSICFG::write_ws_log("| $log(hPA_add): (".join(",",@new_vals).")") if $config::aosi_debug_level > 2;
                                $dummy = 'Added';
                            }
                            elsif($type == 2){ # Modify
                                foreach $cur_val(@{$mod_hash->{$att}}){
                                    push(@new_vals, $cur_val);
                                }
                                AOSICFG::write_ws_log("| $log(hPA_mod): (".join(",",@new_vals).")") if $config::aosi_debug_level > 2;
                                $dummy = 'Modified';
                            }
                            elsif($type == 3){ # delete
                                foreach $old_val(@old_vals){
                                    if($old_val ne $val){
                                        push(@new_vals, $old_val);
                                    }
                                }
                                if($#old_vals > -1 and ($#old_vals != $#new_vals)){
                                    $bin_error = 0;
                                    AOSICFG::write_ws_log("| $log(hPA_del): (".join(",",@new_vals).")") if $config::aosi_debug_level > 2;
                                    $dummy = 'Deleted';
                                }
                                else {
                                    $bin_error = -71;
                                    AOSICFG::write_ws_log("| $log(hPA_del): nothing to delete.") if $config::aosi_debug_level > 2;
                                    $bin_msg = 'Nothing to delete!';
                                }
                            }
                            if($bin_error == 0){
                                AOSICFG::write_ws_log("| $log(hPA_new_vals):(".join(",",@new_vals).")") if $config::aosi_debug_level > 2;
                                $mod_hash->{$att} = \@new_vals;
                                AOSICFG::write_ws_log("| $log(mod_hPA):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                $mesg2 = $ldap->modify( $dn_in,  delete => [$att]);
                                AOSICFG::write_ws_log("| $log(mod_hPA del):code:".$mesg2->code.";error:".$mesg2->error) if $config::aosi_debug_level > 3;
                                $bin_error = $mesg2->code;
                                $bin_msg = $mesg2->error;
                                $mesg2 = $ldap->modify( $dn_in,  replace => $mod_hash);
                                $bin_error = $mesg2->code;
                                $bin_msg = $mesg2->error;
                                AOSICFG::write_ws_log("| $log(mod_hPA replace):code:".$mesg2->code.";error:".$mesg2->error) if $config::aosi_debug_level > 3;
                            }
                            if($bin_error == 0){
                                $out_msg .= '<status code="0">'.$dummy.' attrbiute '.$att.'!</status>';
                            }
                            else {
                                $mod_hash->{$att} = \@old_vals;
                                $mesg2 = $ldap->modify( $dn_in,  replace => $mod_hash);
                                if($mesg2->code != 0){
                                    $bin_error = $mesg2->code;
                                    $bin_msg = $mesg2->error;
                                    $code += abs($bin_error);
                                    $err_num++;
                                    $out_msg .= '<status code="'.$bin_error.'">'.$bin_msg.'</status>';
                                }
                                AOSICFG::write_ws_log("| $log(mod_hPA put old vals back):code:".$mesg2->code.";error:".$mesg2->error) if $config::aosi_debug_level > 3;
                            }
                        }
                        elsif ($att eq 'hrEduPersonPrimaryAffiliation') {
                            if($type == 1){ # Add
                                $out_msg .= '<status code="-41">Attribute '.$att.' cannot be added! Add a value to hrEduPersonAffiliation!</status>';
                                $code = 41;
                                $err_num++;
                            }
                            elsif($type == 3){ # Delete
                                $out_msg .= '<status code="-43">Attribute '.$att.' cannot be deleted! In order to delete it you have to delete hrEduPersonAffiliation!</status>';
                                $code = 43;
                                $err_num++;
                            }
                            elsif($type == 2){ # Modify
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", 'hrEduPersonAffiliation', 0, -1, 0);
                                $bin_xml = $bin_hash->{'result'};
                                $cur_code = $bin_hash->{'code'};
                                ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                if(ref($bin_val) eq 'ARRAY'){
                                    @hep_af_vals = @$bin_val;
                                }
                                else {
                                    @hep_af_vals = ($bin_val);
                                }
                                AOSICFG::write_ws_log("| $log(hep_PAf get hep_Af): code:$cur_code;values:[".join(",",@hep_af_vals)."]") if $config::aosi_debug_level > 2;
                                if($cur_code == 0){
                                    if($#hep_af_vals > -1){
                                        $dummy = 0;
                                        $cur_val = @{$mod_hash->{$att}}[0];
                                        foreach $tmp_val(@hep_af_vals){
                                            if($tmp_val eq $cur_val){
                                                $dummy++;
                                            }
                                        }
                                        AOSICFG::write_ws_log("| $log(hep_PAf find equal in hep_Af): $dummy ($cur_val)") if $config::aosi_debug_level > 2;
                                        if($dummy > 0){
                                            $mesg2 = $ldap->modify( $dn_in,  replace => $mod_hash);
                                            $cur_code = $mesg2->code;
                                            if($cur_code == 0){
                                                $code = $cur_code;
                                                $out_msg .= '<status code="0">Modified attribute '.$att.'!</status>';
                                            }
                                            else {
                                                $out_msg .= '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                                $code += $cur_code;
                                                $err_num++;
                                            }
                                        }
                                        else {
                                            $cur_code = -48;
                                            $out_msg .= '<status code="'.$cur_code.'">hrEduPersonPrimaryAffiliation can be set only to one of the values in hrEduPersonAffiliation!</status>';
                                            $code += abs($cur_code);
                                            $err_num++;
                                        }
                                    }
                                    else {
                                        $cur_code = -47;
                                        $out_msg .= '<status code="'.$cur_code.'">Attribute hrEduPersonAffiliation is empty! Add a value to this attribute first!</status>';
                                        $code += abs($cur_code);
                                        $err_num++;
                                    }
                                }
                                else {
                                    $out_msg .= '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                    $code += $cur_code;
                                    $err_num++;
                                }
                            }
                        }
                        elsif ($att eq 'hrEduPersonAffiliation') {
                            if($type == 1){ # Add
                                AOSICFG::write_ws_log("| $log(add):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                AOSICFG::write_ws_log("| $log(add2):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                $mesg2 = $ldap->modify( $dn_in,  add => $mod_hash);
                                $cur_code = $mesg2->code;
                                if($cur_code == 0){
                                    my $all_attrs = join(", ", keys %$mod_hash);
                                    $out_msg .= '<status code="0">Added attribute '.$all_attrs.'!</status>';
                                }
                                else {
                                    $out_msg .= '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                    $code += abs($cur_code);
                                    $err_num++;
                                }
                            }
                            elsif($type == 3 or $type == 2){ # Delete or Modify
                                $bin_hash = ldapGenSearch($log, $user, $pwd, $base, "(uid=$uid)", 'hrEduPersonPrimaryAffiliation', 0, -1, 0);
                                $bin_xml = $bin_hash->{'result'};
                                $cur_code = $bin_hash->{'code'};
                                ($bin_ent, $bin_attr_ex) = xml2entries($bin_xml);
                                $bin_val = $bin_ent->[0]->{'attrs'}->[0]->{'val'};
                                if(ref($bin_val) eq 'ARRAY'){
                                    $bin_val = $bin_val->[0];
                                }
                                AOSICFG::write_ws_log("| $log(hep_Af get hep_PAf): code:$cur_code;value:$bin_val") if $config::aosi_debug_level > 2;
                                $dummy = 0;
                                @hep_af_vals = @{$mod_hash->{$att}};
                                foreach $tmp_val(@hep_af_vals){
                                    if($tmp_val eq $bin_val){
                                        $dummy++;
                                    }
                                }
                                AOSICFG::write_ws_log("| $log(hep_Af find equal in hep_PAf): $dummy ($bin_val) in [".join(",",@hep_af_vals)."]") if $config::aosi_debug_level > 2;
                                if($type == 3){ # Delete
                                    if($dummy == 0){
                                        AOSICFG::write_ws_log("| $log(hep_Af delete):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                        $mesg2 = $ldap->modify( $dn_in,  delete => $mod_hash);
                                        $cur_code = $mesg2->code;
                                        if($cur_code == 0){
                                            $code = $cur_code;
                                            $out_msg .= '<status code="0">Deleted attribute '.$att.'!</status>';
                                        }
                                        else {
                                            $out_msg .= '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                            $code += $cur_code;
                                            $err_num++;
                                        }
                                    }
                                    else {
                                        $cur_code = -51;
                                        $out_msg .= '<status code="'.$cur_code.'">This value cannot be deleted, hrEduPersonPrimaryAffiliation is set to that value!</status>';
                                        $code += abs($cur_code);
                                        $err_num++;
                                    }
                                }
                                elsif($type == 2){ # Modify
                                    if($dummy == 1){
                                        AOSICFG::write_ws_log("| $log(hep_Af delete):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                        $mesg2 = $ldap->modify( $dn_in,  replace => $mod_hash);
                                        $cur_code = $mesg2->code;
                                        if($cur_code == 0){
                                            $code = $cur_code;
                                            $out_msg .= '<status code="0">Modified attribute '.$att.'!</status>';
                                        }
                                        else {
                                            $out_msg .= '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                            $code += $cur_code;
                                            $err_num++;
                                        }
                                    }
                                    else {
                                        $cur_code = -52;
                                        $out_msg .= '<status code="'.$cur_code.'">This attribute cannot be modified to values which do not contain the value of hrEduPersonPrimaryAffiliation!</status>';
                                        $code += abs($cur_code);
                                        $err_num++;
                                    }
                                }
                            }
                        }
                        else {
                            if($type == 1){ # Add
                                AOSICFG::write_ws_log("| $log(add):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                $mesg2 = $ldap->modify( $dn_in,  add => $mod_hash);
                                $cur_code = $mesg2->code;
                                if($cur_code == 0){
                                    my $all_attrs = join(", ", keys %$mod_hash);
                                    $out_msg .= '<status code="0">Added attribute '.$all_attrs.'!</status>';
                                }
                                else {
                                    $out_msg .= '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                    $code += abs($cur_code);
                                    $err_num++;
                                }
                            }
                            elsif($type == 2){ # Modify
                                AOSICFG::write_ws_log("| $log(mod):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                $mesg2 = $ldap->modify( $dn_in,  replace => $mod_hash);
                                $cur_code = $mesg2->code;
                                if($cur_code == 0){
                                    my $all_attrs = join(", ", keys %$mod_hash);
                                    $out_msg .= '<status code="0">Modified attribute '.$all_attrs.'!</status>';
                                }
                                else {
                                    $out_msg .= '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                    $code += abs($cur_code);
                                    $err_num++;
                                }
                            }
                            elsif($type == 3){ # delete
                                AOSICFG::write_ws_log("| $log(del):".join(",", @del)) if $config::aosi_debug_level > 2;
                                my($all_attrs);
                                if($#del > -1){
                                    $all_attrs = join(", ", @del);
                                    $mesg2 = $ldap->modify( $dn_in,  delete => \@del);
                                }
                                @mod_keys = keys %$mod_hash;
                                if($#mod_keys > -1){
                                    $all_attrs .= ','.join(", ", keys %$mod_hash);
                                    $mesg2 = $ldap->modify( $dn_in,  delete => $mod_hash);
                                }
                                AOSICFG::write_ws_log("| $log(del_dumphash):".AOSICFG::dump_hash($mod_hash)) if $config::aosi_debug_level > 2;
                                $cur_code = $mesg2->code;
                                if($cur_code == 0){
                                    $out_msg .= '<status code="0">Deleted attribute '.$all_attrs.'!</status>';
                                }
                                else {
                                    $out_msg .= '<status code="'.$cur_code.'">'.$mesg2->error.'</status>';
                                    $code += abs($cur_code);
                                    $err_num++;
                                }
                            }
                        }
                    }
                    if($mesg2->code == 0){
                        AOSICFG::write_ws_log("| $log(final bin_error):".$bin_error) if $config::aosi_debug_level > 2;
                        if($bin_error == 0){
                            $cur_code = $mesg2->code;
                            $code += abs($cur_code);
                            #$hash->{'result'} = $out_msg;
                        }
                        else {
                            $cur_code = $bin_error;
                            $err_msg = $bin_msg;
                            $out_msg .= '<status code="'.$cur_code.'">'.$err_msg.'</status>';
                            $code += abs($cur_code);
                            $err_num++;
                        }
                    }
                    else {
                        # Ovo makni
                        #$cur_code = $mesg2->code;
                        #$err_msg = $mesg2->error;
                        #$out_msg .= '<status BBB code="'.$cur_code.'">'.$err_msg.'</status>';
                        #$code += abs($cur_code);
                        #$err_num++;
                    }
                }
                else {
                    $code = 83;
                    $out_msg .= '<status code="-'.$code.'">You are not allowed to alter attributes for this user!</status>';
                    $err_num++;
                }
                $out_msg .= '</entry>';

            }
            $hash->{'code'} = $code;
            $hash->{'error'} = $err_num.' error(s) occured!';
            $hash->{'result'} = '<?xml version="1.0"?><ldap action="'.$service.'" version="'.$config::aosi_version.'">'.$out_msg.'</ldap>';;
            $mesg2 = $ldap->unbind;  # take down session

    }
    else {
        $hash->{'code'} = $auth_code;
        $hash->{'error'} = $auth_error;
        $hash->{'result'} = '<?xml version="1.0"?><ldap action="'.$service.'" version="'.$config::aosi_version.'"><entry dn="'.$dn.'"><status code="'.$auth_code.'">'.$auth_error.'</status></entry></ldap>';
    }
    AOSICFG::write_ws_log("| $log(out):".AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 2;
    return $hash;
}

######################################################################
#
######################################################################
sub ck_auth {
    my($ldap, $dn, $pwd, $ck_admin, $ldap_error) = @_;
    my($adm, $admin, $admin_type, $attrib, $base, $code, $dc, $dec_pwd);
    my($domain, $error, $full_uid, $hredu_uid, $is_admin_ldap, $mesg);
    my($rest, $tld, $type, $user, %options, @entries, $entry, $log);
    my($out_filter, $val, $dummy, @urn, $urn_val, $urn_key, @or, @keys, %and, $temp);
    my(@entries2, $entry2, @values2);
    $log = 'ck_auth';

    $is_admin_ldap = 0;
    $out_filter = '';
    AOSICFG::write_ws_log("|/ $log(in): dn=$dn;ck_admin=$ck_admin;ldap=$ldap;ip=$AOSI::client_ip;use_valid_ip:".$config::use_valid_ip);
    if($ldap){
        if($dn eq '' || $pwd eq ''){
            AOSICFG::write_ws_log("|| $log: No DN or password") if $config::aosi_debug_level > 3;
            $type = 'IP';
            if($config::use_valid_ip == 1){
                AOSICFG::write_ws_log("|| $log(no_dn_pwd): using valid ip") if $config::aosi_debug_level > 3;
                if(AOSICFG::ip_is_valid($AOSI::client_ip)){
                    AOSICFG::write_ws_log("|| $log(no_dn_pwd):using valid ip=valid") if $config::aosi_debug_level > 3;
                    $is_admin_ldap = 1;
                    $code = 0;
                }
                else {
                    $code = -10;
                    $error = 'No DN or password from invalid IP: '.$AOSI::client_ip.'!';
                    AOSICFG::write_ws_log("|| $log(no_dn_pwd):using valid ip=INVALID") if $config::aosi_debug_level > 3;
                }
            }
            else {
                if($dn eq ''){
                    $code = -12;
                    $error = 'DN is not specified!';
                    AOSICFG::write_ws_log("|| $log(no_dn_pwd): No DN") if $config::aosi_debug_level > 3;
                }
                elsif($pwd eq ''){
                    $code = -13;
                    $error = 'No password specified!';
                    AOSICFG::write_ws_log("|| $log(no_dn_pwd): No password") if $config::aosi_debug_level > 3;
                }
            }
        }
        else {
            $type = 'user/pwd';
            $dec_pwd = decode_base64($pwd);
            $mesg = $ldap->bind($dn, password => $dec_pwd, version => 3);
            $code = $mesg->code;
            $error = $mesg->error;
            AOSICFG::write_ws_log("|| $log(step one - bind):$dn;code:$code;error:$error") if $config::aosi_debug_level > 2;
            if($code == 0){ # Second auth step
                $attrib = 'hrEduPersonUniqueID';  #'hrEduPersonUniqueID'
                ($user, $domain, $tld) = split(/,/, $dn);
                ($rest, $user) = split(/=/, $user);
                $base = "$domain,$tld";
                ($dc, $domain) = split(/=/, $domain);
                ($dc, $tld) = split(/=/, $tld);
                $full_uid = $user.'@'.$domain.'.'.$tld;
                %options = (base   => $base,
              filter => "(uid=$user)",
              #sizelimit => 0,
              timelimit => 180,
              attrs => [$attrib]);
                $mesg = $ldap->search(%options);
                $code = $mesg->code;
                $error = $mesg->error;
                @entries = $mesg->entries;
                $hredu_uid = $entries[0]->get_value($attrib);
    
                #($hredu_uid, $rest) = split(/@/, $hredu_uid);
                AOSICFG::write_ws_log("|| $log(step two - ck_uid):$hredu_uid ? $full_uid;code:$code;error:$error") if $config::aosi_debug_level > 2;
                if($hredu_uid eq $full_uid){
                    $code = 0;
                }
                else {
                    $code = -11;
                    $error = 'hrEduPersonUniqueID($hredu_uid) does not match $full_uid!';
                }
            }
            if($code == 0 and $ck_admin == 1){
                $out_filter = '';
                ($user, $rest) = split(/,/, $dn);
                ($rest, $user) = split(/=/, $user);
                $admin = AOSICFG::user_is_valid($user);
                if($admin > 0){
                    $is_admin_ldap = 1;
                    $admin_type = "file";
                    AOSICFG::write_ws_log("|| $log(admin check - file): OK") if $config::aosi_debug_level > 2;
                }
                else {
                    $attrib = 'hrEduPersonRole';
                    %options = (base   => $base,
                      filter => "(uid=$user)",
                      #sizelimit => 0,
                      timelimit => 60,
                      attrs => [$attrib]);
                    $mesg = $ldap->search(%options);
                    $code = $mesg->code;
                    $error = $mesg->error;
                    AOSICFG::write_ws_log("|| $log(search for hrEduPersonRole):code:$code;error:$error") if $config::aosi_debug_level > 2;
                    @entries = $mesg->entries;
                    AOSICFG::write_ws_log("|| $log(search for hrEduPersonRole):entries:".$#entries) if $config::aosi_debug_level > 2;
                    $entry = $entries[0];
                    @values = @{$entry->get_value( $attrib, asref => 1 )};
                    LOOP:
                    foreach $val(@values){
                        AOSICFG::write_ws_log("|| $log(hrEduPersonRole val):$val") if $config::aosi_debug_level > 3;
                        my %and;
                        if($val =~ /^urn\:*/){
                            AOSICFG::write_ws_log("|| $log(hrEduPersonRole val): has urn") if $config::aosi_debug_level > 3;
                            $is_admin_ldap = 3;
                            $admin_type = "hrEduPersonRole=urn";
                            @urn = split(/:/, $val);
                            $dummy = shift(@urn);
    					    $dummy = shift(@urn);
    					    if($#urn > -1){
        						while($urn_key = shift(@urn)){
        							$urn_val = shift(@urn);
        							if($urn_val ne ''){
                                        push(@{$and{$urn_key}}, $urn_val);
    				                }
        						}
        					}
        					else {
        						@or = ();
        						last LOOP;
        					}
        					push(@or, \%and);
                        }
                        if($val eq 'administrator imenika'){
                            AOSICFG::write_ws_log("|| $log(hrEduPersonRole val): has urn") if $config::aosi_debug_level > 3;
                            $is_admin_ldap = 2;
                            $admin_type = "hrEduPersonRole";
                            @or = ();
                            last LOOP;
                        }
                    }
                    if($#or > -1){
                        foreach $val(@or){
                            $dummy = '';
                            @keys = keys %$val;
                            foreach $urn_key(@keys){
                                @urn = @{$val->{$urn_key}};
                                if($#urn > 0){
                                    $dummy .= '(|';
                                    foreach $urn_val(@urn){
                                        $dummy .= '('.$urn_key.'='.$urn_val.')';
                                    }
                                    $dummy .= ')';
                                }
                                else {
                                    $dummy .= '('.$urn_key.'='.$urn[0].')';
                                }
                            }
                            if($dummy ne ''){
                                $out_filter .= '(&'.$dummy.')';
                            }
                        }
                        $out_filter = '(|'.$out_filter.')';
                    }

                   if ($is_admin_ldap == 0){
                        $admin_type = "none";
                    }
                    AOSICFG::write_ws_log("|| $log(admin check - $attrib): OK") if $config::aosi_debug_level > 2;
                }
                $type .= " (admin: $admin_type)";
            }
        }
    }
    else {
        $code = -18;
        $error = $ldap_error;
    }
    AOSICFG::write_ws_log("|\\ $log(out):type:$type;code:$code;error:$error;out_filter:$out_filter");

    return $code, $error, $is_admin_ldap, $out_filter, \@or;
}


1;
